/****************************************************************************
 **
 ** Copyright (C) 2001-2005 Tilo Riemer <riemer@lincvs.org> and
 **                         Frank Hemer <frank@hemer.org>
 **
 **
 **----------------------------------------------------------------------------
 **
 **----------------------------------------------------------------------------
 **
 ** LinCVS is available under two different licenses:
 **
 ** If LinCVS is linked against the GPLed version of Qt 
 ** LinCVS is released under the terms of GPL also.
 **
 ** If LinCVS is linked against a nonGPLed version of Qt 
 ** LinCVS is released under the terms of the 
 ** LinCVS License for non-Unix platforms (LLNU)
 **
 **
 ** LinCVS License for non-Unix platforms (LLNU):
 **
 ** Redistribution and use in binary form, without modification, 
 ** are permitted provided that the following conditions are met:
 **
 ** 1. Redistributions in binary form must reproduce the above copyright
 **    notice, this list of conditions and the following disclaimer in the
 **    documentation and/or other materials provided with the distribution.
 ** 2. It is not permitted to distribute the binary package under a name
 **    different than LinCVS.
 ** 3. The name of the authors may not be used to endorse or promote
 **    products derived from this software without specific prior written
 **    permission.
 ** 4. The source code is the creative property of the authors.
 **    Extensions and development under the terms of the Gnu Public License
 **    are limited to the Unix platform. Any distribution or compilation of 
 **    the source code against libraries licensed other than gpl requires 
 **    the written permission of the authors.
 **
 **
 ** THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR 
 ** IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 ** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 ** ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY 
 ** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 ** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
 ** GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 ** INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 ** WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 ** NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
 ** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 **
 **
 **
 ** LinCVS License for Unix platforms:
 **
 ** This program is free software; you can redistribute it and/or modify 
 ** it under the terms of the GNU General Public License as published by 
 ** the Free Software Foundation; either version 2 of the License, or 
 ** (at your  option) any later version.  This program is distributed in 
 ** the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 ** even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 ** PARTICULAR PURPOSE.
 **
 ** See the GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program; if not, write to the Free Software Foundation,
 ** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 **
 *****************************************************************************/

#include "Tag.h"

#include <qapplication.h>
#include <assert.h>

#include "globals.h"
#include "TagDialogImpl.h"
#include "TagStdDialogImpl.h"

Tag * Tag::tagFile(const QIconSet &whatsThisIconSet, QWidget* parent, CvsDirListView * workBench, QStringList * tagList)
{
   return setup(whatsThisIconSet, parent, workBench, false, tagList);
}

Tag * Tag::tagDir(const QIconSet &whatsThisIconSet, QWidget* parent, CvsDirListView * workBench, QStringList * tagList)
{
   return setup(whatsThisIconSet, parent, workBench, true, tagList);
}

Tag * Tag::setup(const QIconSet &whatsThisIconSet, QWidget* parent, CvsDirListView * workBench, bool isDir, QStringList * tagList) {

   DirBase * dir = workBench->selectedItem();
   if (!dir) return NULL;

   if (dir->getType() == DirBase::Cvs) {

      QDialog * dlg = 0;
      QString topModule = dir->topControlledDir()->relativeName();

      if (LookAndFeel::g_stdDlg) {
	 dlg = new TagStdDialogImpl( (isDir) ? dir->repository() : QString::null,
	       tagList, topModule,
	       whatsThisIconSet,
	       parent, (isDir) ? tr("Tag a Directory") : tr("Tag File(s)") );
      } else {
	 dlg = new TagDialogImpl( (isDir) ? dir->repository() : QString::null,
	       tagList, topModule,
	       whatsThisIconSet,
	       parent, (isDir) ? tr("Tag a Directory") : tr("Tag File(s)") );
      }

      if (dlg->exec()) {
	 Tag * p = new Tag(whatsThisIconSet, parent, workBench, dir, isDir, tagList);
	 p->m_dlg = dlg;
	 p->acceptCvs();
	 return p;
      }

   } else {
      qDebug("Tag: Method not implemented");
   }

   return NULL;
}

Tag::Tag(const QIconSet &whatsThisIconSet, QWidget* parent, CvsDirListView * workBench, DirBase * dir, bool isDir, QStringList * tagList)
   :  m_whatsThisIconSet(whatsThisIconSet),
      m_parent(parent),
      m_workBench(workBench),
      m_dir(dir),
      m_dlg(NULL),
      m_isDir(isDir),
      m_currentTagList(tagList)
{
   connect(this,SIGNAL(deleteObject(QObject *)),parent,SLOT(slot_deleteObject(QObject *)));
   connect(this,SIGNAL(checkInProgress(bool)),parent,SLOT(checkInProgress(bool)));
   connect(this,SIGNAL(checkTree(DirBase *)),parent,SLOT(slot_checkStatusOfTree(DirBase *)));
}

Tag::~Tag() {
   if (m_dlg) delete m_dlg;
   m_dlg = NULL;
}

void Tag::acceptCvs() {

   assert(m_dir);
   assert(m_dlg);

   QString command = "";
   if (LookAndFeel::g_stdDlg) {
      TagStdDialogImpl * dlg = static_cast<TagStdDialogImpl *>(m_dlg);
      assert(dlg);
      command = dlg->getCommand();
   } else {
      TagDialogImpl * dlg = static_cast<TagDialogImpl *>(m_dlg);
      assert(dlg);
      command = dlg->getCommand();
   }

   delete m_dlg;
   m_dlg = NULL;

   QString files;
   if (!m_isDir) {
      QStringList filesList = m_dir->getSelectedFiles();
      if (filesList.isEmpty()) {
         reject();
         return;
      }
      masqQuoteMarks(&filesList);
      files = masqWs(filesList.join("\" \""));
   }

   QString dir = m_dir->fullName();
   QString topModule = m_dir->topControlledDir()->relativeName();
   
   callInteractive( topModule, dir, command,
	 files, CVS_TAG_CMD,
	 ExtApps::g_cvsRsh.path,  //additional options of cvsRsh not supported yet
	 NOROOT);
}

void Tag::reject() {
   emit deleteObject(this);
}

void Tag::cvsCallStarted() {
   QApplication::setOverrideCursor(Qt::waitCursor);
}

void Tag::cvsCallFinished() {
   QApplication::restoreOverrideCursor();
}

void Tag::afterCall(int cmd,CvsBuffer*,bool failed) {
   cvsCallFinished();
   if (failed) {
      reject();
      return;
   }
   switch( cmd) {
      case CVS_TAG_CMD: {
	 emit checkInProgress(true);
	 m_dir->postCallCheck(DirBase::Controled|(m_isDir ? DirBase::Tree : 0));
	 emit checkInProgress(false);
	 break;
      }
      default: {
	 qDebug("Tag::afterCall: Unknown cmd");
      }
   }
   reject();
}
